from .agent import SingleRoomAgent
from meshagent.api import websocket_protocol, RoomClient
import asyncio
import signal


async def connect_development_agent(*, room_name: str, agent: SingleRoomAgent):
    async with RoomClient(
        protocol=websocket_protocol(
            participant_name=agent.name, room_name=room_name, role="agent"
        )
    ) as room:
        await agent.start(room=room)

        try:
            term = asyncio.Future()

            def clean_termination(signal, frame):
                term.set_result(True)

            signal.signal(signal.SIGTERM, clean_termination)
            signal.signal(signal.SIGABRT, clean_termination)

            await asyncio.wait(
                [asyncio.ensure_future(room.protocol.wait_for_close()), term],
                return_when=asyncio.FIRST_COMPLETED,
            )

        finally:
            await agent.stop()



from typing import Optional
from copy import deepcopy
from meshagent.api import RoomClient
from meshagent.tools.toolkit import Toolkit
from meshagent.api.participant import Participant

import uuid


class AgentChatContext:
    def __init__(
        self,
        *,
        messages: Optional[list[dict]] = None,
        system_role: Optional[str] = None,
        previous_messages: Optional[list[dict]] = None,
        previous_response_id: Optional[str] = None,
    ):
        self.id = str(uuid.uuid4())
        if messages is None:
            messages = list[dict]()
        self._messages = messages.copy()
        if system_role is None:
            system_role = "system"
        self._system_role = system_role

        if previous_messages is None:
            previous_messages = list[dict]()

        self._previous_response_id = previous_response_id
        self._previous_messages = previous_messages

    @property
    def messages(self):
        return self._messages

    @property
    def system_role(self):
        return self._system_role

    @property
    def previous_messages(self):
        return self._previous_messages

    @property
    def previous_response_id(self):
        return self._previous_response_id

    def track_response(self, id: str):
        self._previous_response_id = id
        self._previous_messages.extend(self.messages)
        self.messages.clear()

    def append_rules(self, rules: list[str]):
        system_message = None

        for m in self.messages:
            if m["role"] == self.system_role:
                system_message = m
                break

        if system_message is None:
            system_message = {"role": self.system_role, "content": ""}
            self.messages.insert(0, system_message)

        plan = f"Rules:\n-{'\n-'.join(rules)}\n"
        system_message["content"] = system_message["content"] + plan

    def append_assistant_message(self, message: str) -> None:
        self.messages.append({"role": "assistant", "content": message})

    def append_user_message(self, message: str) -> None:
        self.messages.append({"role": "user", "content": message})

    def append_user_image(self, url: str) -> None:
        self.messages.append(
            {
                "role": "user",
                "content": [
                    {"type": "image_url", "image_url": {"url": url, "detail": "auto"}}
                ],
            }
        )

    def copy(self) -> "AgentChatContext":
        return AgentChatContext(
            messages=deepcopy(self.messages), system_role=self._system_role
        )

    def to_json(self) -> dict:
        return {
            "messages": self.messages,
            "system_role": self.system_role,
            "previous_messages": self.previous_messages,
            "previous_response_id": self.previous_response_id,
        }

    @staticmethod
    def from_json(json: dict):
        return AgentChatContext(
            messages=json["messages"],
            system_role=json.get("system_role", None),
            previous_messages=json.get("previous_messages", None),
            previous_response_id=json.get("previous_response_id", None),
        )


class AgentCallContext:
    def __init__(
        self,
        *,
        chat: AgentChatContext,
        room: RoomClient,
        toolkits: Optional[list[Toolkit]] = None,
        caller: Optional[Participant] = None,
        on_behalf_of: Optional[Participant] = None,
    ):
        self._room = room
        if toolkits is None:
            toolkits = list[Toolkit]()
        self._toolkits = toolkits
        self._chat = chat
        self._caller = caller
        self._on_behalf_of = on_behalf_of

    @property
    def toolkits(self):
        return self._toolkits

    @property
    def chat(self):
        return self._chat

    @property
    def caller(self):
        return self._caller

    @property
    def on_behalf_of(self):
        return self._on_behalf_of

    @property
    def room(self):
        return self._room



from meshagent.api import RoomClient, Requirement, RoomException, Participant
from meshagent.tools import validate_openai_schema
import json
from typing import Optional
import asyncio


async def generate_json(
    *,
    on_behalf_of: Optional[Participant] = None,
    room: RoomClient,
    prompt: Optional[str] = None,
    output_schema: dict,
    requires: Optional[list[Requirement]] = None,
) -> dict:
    # make sure agent is in the room before proceeding
    agent = None
    for i in range(10):
        agents = await room.agents.list_agents()
        for a in agents:
            if a.name == "meshagent.schema_planner":
                agent = a
                break

        if agent is not None:
            break

        await asyncio.sleep(1)

    if agent is None:
        raise RoomException(
            "unable to locate required agent (meshagent.schema_planner)"
        )

    if prompt is None:
        prompt = f"ask me a series of questions to completely fill out the data structure described by this JSON schema ${json.dumps(output_schema)}. If you need to ask multiple questions, try to include all of them in a single form."

    validate_openai_schema(output_schema)
    return await room.agents.ask(
        on_behalf_of=on_behalf_of,
        agent="meshagent.schema_planner",
        requires=requires,
        arguments={
            "prompt": prompt,
            "output_schema": output_schema,
        },
    )



from meshagent.agents import TaskRunner, RequiredToolkit, SingleRoomAgent
from meshagent.tools import Toolkit, Tool, ToolContext
from meshagent.api.room_server_client import (
    TextDataType,
    VectorDataType,
    FloatDataType,
    IntDataType,
)
from openai import AsyncOpenAI
from typing import Optional
from meshagent.api.chan import Chan

import hashlib
import chonkie
import asyncio
import logging

from functools import wraps

import os

# TODO: install chonkie, chonkie[semantic], openai


def _async_debounce(wait):
    def decorator(func):
        task = None

        @wraps(func)
        async def debounced(*args, **kwargs):
            nonlocal task

            async def call_func():
                await asyncio.sleep(wait)
                await func(*args, **kwargs)

            if task and not task.done():
                task.cancel()

            task = asyncio.create_task(call_func())
            return task

        return debounced

    return decorator


logger = logging.getLogger("indexer")


class Chunk:
    def __init__(self, *, text: str, start: int, end: int):
        self.text = text
        self.start = start
        self.end = end


class Chunker:
    async def chunk(
        self, *, text: str, max_length: Optional[int] = None
    ) -> list[Chunk]:
        pass


class ChonkieChunker(Chunker):
    def __init__(self, chunker: Optional[chonkie.BaseChunker] = None):
        super().__init__()

        if chunker is None:
            chunker = chonkie.SemanticChunker()

        self._chunker = chunker

    async def chunk(
        self, *, text: str, max_length: Optional[int] = None
    ) -> list[Chunk]:
        chunks = await asyncio.to_thread(self._chunker.chunk, text=text)
        mapped = []
        for chunk in chunks:
            mapped.append(
                Chunk(text=chunk.text, start=chunk.start_index, end=chunk.end_index)
            )
        return mapped


class Embedder:
    def __init__(self, *, size: int, max_length: int):
        self.size = size
        self.max_length = max_length

    async def embed(self, *, text: str) -> list[float]:
        pass


class OpenAIEmbedder(Embedder):
    def __init__(
        self,
        *,
        size: int,
        max_length: int,
        model: str,
        openai: Optional[AsyncOpenAI] = None,
    ):
        if openai is None:
            openai = AsyncOpenAI()

        self._openai = openai
        self._model = model

        super().__init__(size=size, max_length=max_length)

    async def embed(self, *, text):
        return (
            (
                await self._openai.embeddings.create(
                    input=text, model=self._model, encoding_format="float"
                )
            )
            .data[0]
            .embedding
        )


class RagTool(Tool):
    def __init__(
        self,
        *,
        name="rag_search",
        table: str,
        title="RAG search",
        description="perform a RAG search",
        rules=None,
        thumbnail_url=None,
        embedder: Optional[Embedder] = None,
    ):
        self.table = table

        super().__init__(
            name=name,
            input_schema={
                "type": "object",
                "additionalProperties": False,
                "required": ["query"],
                "properties": {"query": {"type": "string"}},
            },
            title=title,
            description=description,
            rules=rules,
            thumbnail_url=thumbnail_url,
        )

        self._embedder = embedder

    async def execute(self, context: ToolContext, query: str):
        if self._embedder is None:
            results = await context.room.database.search(
                table=self.table, text=query, limit=10
            )
        else:
            embedding = await self._embedder.embed(text=query)
            results = await context.room.database.search(
                table=self.table, text=query, vector=embedding, limit=10
            )

        results = list(map(lambda r: f"from {r['url']}: {r['text']}", results))

        return {"results": results}


def open_ai_embedding_3_small():
    return OpenAIEmbedder(model="text-embedding-3-small", max_length=8191, size=1536)


def open_ai_embedding_3_large():
    return OpenAIEmbedder(model="text-embedding-3-large", max_length=8191, size=3072)


def open_ai_embedding_ada_2():
    return OpenAIEmbedder(model="text-embedding-ada-002", max_length=8191, size=1536)


class RagToolkit(Toolkit):
    def __init__(self, table: str, embedder: Optional[Embedder] = None):
        if embedder is None:
            embedder = open_ai_embedding_3_large()

        super().__init__(
            name="meshagent.rag",
            title="RAG",
            description="Searches against an index",
            tools=[RagTool(table=table, embedder=embedder)],
        )


class FileIndexEvent:
    def __init__(self, *, path: str, deleted: bool):
        self.path = path
        self.deleted = deleted


class StorageIndexer(SingleRoomAgent):
    def __init__(
        self,
        *,
        name,
        title=None,
        description=None,
        requires=None,
        labels=None,
        chunker: Optional[Chunker] = None,
        embedder: Optional[Embedder] = None,
        table: str = "storage_index",
    ):
        super().__init__(
            name=name,
            title=title,
            description=description,
            requires=requires,
            labels=labels,
        )

        self._chan = Chan[FileIndexEvent]()

        if chunker is None:
            chunker = ChonkieChunker()

        if embedder is None:
            embedder = open_ai_embedding_3_large()

        self.chunker = chunker
        self.embedder = embedder
        self.table = table
        self._vector_index_created = False
        self._fts_created = False

    async def read_file(self, *, path: str) -> str | None:
        pass

    @_async_debounce(10)
    async def refresh_index(self):
        self.room.developer.log_nowait(type="indexer.rebuild", data={})

        indexes = await self.room.database.list_indexes(table=self.table)

        logger.info(f"existing indexes {indexes}")

        for index in indexes:
            if "embedding" in index["columns"]:
                self._vector_index_created = True

            if "text" in index["columns"]:
                self._fts_created = True

        if not self._vector_index_created:
            try:
                logger.info("attempting to create embedding index")
                await self.room.database.create_vector_index(
                    table=self.table, column="embedding", replace=False
                )
                self._vector_index_created = True
            except Exception:
                # Will fail if there aren't enough rows
                pass

        if not self._fts_created:
            try:
                logger.info("attempting to create fts index")
                await self.room.database.create_full_text_search_index(
                    table=self.table, column="text", replace=False
                )
                self._fts_created = True
            except Exception:
                # Will fail if there aren't enough rows
                pass

        if self._fts_created or self._vector_index_created:
            logger.info("optimizing existing index")
            await self.room.database.optimize(table=self.table)

    async def start(self, *, room):
        await super().start(room=room)

        room.storage.on("file.updated", self._on_file_updated)
        room.storage.on("file.deleted", self._on_file_deleted)

        await room.database.create_table_with_schema(
            name=self.table,
            schema={
                "url": TextDataType(),
                "text": TextDataType(),
                "embedding": VectorDataType(
                    size=self.embedder.size, element_type=FloatDataType()
                ),
                "sha": TextDataType(),
            },
            mode="create_if_not_exists",
            data=None,
        )

        def index_task(task: asyncio.Task):
            try:
                task.result()
            except Exception as e:
                logger.error("Index task failed", exc_info=e)

        self._index_task = asyncio.create_task(self._indexer())
        self._index_task.add_done_callback(index_task)

    async def stop(self):
        await super().stop()
        await self._chan.close()

    async def _indexer(self):
        async for e in self._chan:
            try:
                if e.deleted:
                    # todo: consider using sql_alchemy or a library to do the escaping
                    def escape_sql_string(value):
                        if not isinstance(value, str):
                            raise TypeError("Input must be a string")
                        return value.replace("'", "''")

                    self.room.developer.log_nowait(
                        type="indexer.delete", data={"path": e.path}
                    )
                    await self.room.database.delete(
                        table=self.table, where=f"url='{escape_sql_string(e.path)}'"
                    )

                else:
                    self.room.developer.log_nowait(
                        type="indexer.index", data={"path": e.path}
                    )

                    async def lookup_or_embed(*, sha: str, text: str) -> list[float]:
                        # if we already indexed this chunk, lets use the existing embedding instead of generating a new one
                        results = await self.room.database.search(
                            table=self.table,
                            where={
                                "sha": sha,
                            },
                            limit=1,
                        )

                        if len(results) != 0:
                            logger.info(
                                f"chunk found from {e.path} {sha}, reusing embedding"
                            )
                            return results[0]["embedding"]

                        logger.info(
                            f"chunk not found from {e.path} {sha}, generating embedding"
                        )

                        return await self.embedder.embed(text=text)

                    basename = os.path.basename(e.path)

                    chunk_sha = hashlib.sha256(basename.encode("utf-8")).hexdigest()

                    rows = []
                    # let's make the filename it's own chunk
                    rows.append(
                        {
                            "url": e.path,
                            "text": basename,
                            "sha": chunk_sha,
                            "embedding": await lookup_or_embed(
                                sha=chunk_sha, text=basename
                            ),
                        }
                    )

                    text = await self.read_file(path=e.path)
                    if text is not None:
                        # the content will be transformed into additional chunks
                        for chunk in await self.chunker.chunk(
                            text=text, max_length=self.embedder.max_length
                        ):
                            logger.info(
                                f"processing chunk from {e.path}: {chunk.start}"
                            )
                            chunk_sha = hashlib.sha256(
                                chunk.text.encode("utf-8")
                            ).hexdigest()
                            rows.append(
                                {
                                    "url": e.path,
                                    "text": chunk.text,
                                    "embedding": await lookup_or_embed(
                                        sha=chunk_sha, text=chunk.text
                                    ),
                                    "sha": chunk_sha,
                                }
                            )
                    await self.room.database.merge(
                        table=self.table, on="sha", records=rows
                    )
                    await self.refresh_index()

            except Exception as e:
                logger.error("error while indexing", exc_info=e)

    def _on_file_deleted(self, path: str, participant_id: str):
        self._chan.send_nowait(FileIndexEvent(path=path, deleted=True))

    def _on_file_updated(self, path: str, participant_id: str):
        self._chan.send_nowait(FileIndexEvent(path=path, deleted=False))


class SiteIndexer(TaskRunner):
    def __init__(
        self,
        *,
        name,
        chunker: Optional[Chunker] = None,
        embedder: Optional[Embedder] = None,
        title=None,
        description=None,
        requires=None,
        supports_tools=None,
        labels: Optional[list[str]] = None,
    ):
        if chunker is None:
            chunker = ChonkieChunker()

        if embedder is None:
            embedder = open_ai_embedding_3_large()

        self.chunker = chunker
        self.embedder = embedder

        super().__init__(
            name=name,
            title=title,
            description=description,
            requires=[
                RequiredToolkit(name="meshagent.firecrawl", tools=["firecrawl_queue"]),
            ],
            supports_tools=supports_tools,
            input_schema={
                "type": "object",
                "required": ["queue", "table", "url"],
                "additionalProperties": False,
                "properties": {
                    "queue": {"type": "string", "description": "default: firecrawl"},
                    "table": {"type": "string", "description": "default: index"},
                    "url": {"type": "string", "description": "default: index"},
                },
            },
            output_schema={
                "type": "object",
                "required": [],
                "additionalProperties": False,
                "properties": {},
            },
            labels=labels,
        )

    async def ask(self, *, context, arguments):
        queue = arguments["queue"]
        table = arguments["table"]
        url = arguments["url"]

        tables = await context.room.database.list_tables()

        exists = False
        try:
            exists = tables.index(table)
        except ValueError:
            pass

        async def lookup_or_embed(*, sha: str, text: str) -> list[float]:
            # if we already indexed this chunk, lets use the existing embedding instead of generating a new one
            if exists:
                results = await self.room.database.search(
                    table=self.table,
                    where={
                        "sha": sha,
                    },
                    limit=1,
                )

                if len(results) != 0:
                    logger.info(f"chunk found from {url} {sha}, reusing embedding")
                    return results[0]["embedding"]

            logger.info(f"chunk not found from {url} {sha}, generating embedding")

            return await self.embedder.embed(text=text)

        async def crawl():
            logger.info(f"starting to crawl: {url}")
            await context.room.agents.invoke_tool(
                toolkit="meshagent.firecrawl",
                tool="firecrawl_queue",
                arguments={"url": url, "queue": queue, "limit": 100},
            )

            logger.info(f"done with crawl: {url}")
            await context.room.queues.send(name=queue, message={"done": True})

        def crawl_done(task: asyncio.Task):
            try:
                task.result()
            except Exception as e:
                logger.error("crawl failed", exc_info=e)

        crawl_task = asyncio.create_task(crawl())
        crawl_task.add_done_callback(crawl_done)

        rows = []

        id = 0

        while True:
            message = await context.room.queues.receive(
                name=queue, create=True, wait=True
            )

            if message is None:
                break

            if message.get("type", None) == "crawl.completed":
                break

            if "data" in message:
                for data in message["data"]:
                    try:
                        url: str = data["metadata"]["url"]
                        text: str = data["markdown"]
                        title: str = data["metadata"]["title"]
                        title_sha: str = hashlib.sha256(
                            text.encode("utf-8")
                        ).hexdigest()

                        logger.info(f"processing crawled page: {url}")

                        # let's make the title it's own chunk
                        rows.append(
                            {
                                "id": id,
                                "url": url,
                                "text": title,
                                "sha": title_sha,
                                "embedding": await lookup_or_embed(
                                    sha=title_sha, text=title
                                ),
                            }
                        )

                        id = id + 1

                        # the content will be transformed into additional chunks
                        for chunk in await self.chunker.chunk(
                            text=text, max_length=self.embedder.max_length
                        ):
                            logger.info(f"processing chunk from {url}: {chunk.text}")
                            chunk_sha = hashlib.sha256(
                                chunk.text.encode("utf-8")
                            ).hexdigest()
                            rows.append(
                                {
                                    "id": id,
                                    "url": url,
                                    "text": chunk.text,
                                    "embedding": await lookup_or_embed(
                                        sha=chunk_sha, text=chunk.text
                                    ),
                                }
                            )

                            id = id + 1

                    except Exception as e:
                        logger.error(f"failed to process: {url}", exc_info=e)

        logger.info(f"saving crawl: {url}")

        await context.room.database.create_table_with_schema(
            name=table,
            schema={
                "id": IntDataType(),
                "url": TextDataType(),
                "text": TextDataType(),
                "embedding": VectorDataType(
                    size=self.embedder.size, element_type=FloatDataType()
                ),
                "sha": TextDataType(),
            },
            mode="overwrite",
            data=rows,
        )

        if len(rows) > 255:
            await context.room.database.create_vector_index(
                table=table, column="embedding"
            )

        await context.room.database.create_full_text_search_index(
            table=table, column="text"
        )

        return {}



from meshagent.agents.worker import Worker
from meshagent.api.room_server_client import TextDataType
from email import message_from_bytes
from email.message import EmailMessage
from email.policy import default
from meshagent.tools import ToolContext, Toolkit
import email.utils
from meshagent.api import ParticipantToken
from meshagent.api import RoomClient
from meshagent.agents import AgentChatContext
from datetime import datetime, timezone

import secrets

from typing import Literal, Optional
import json

import uuid
import logging

import os
import aiosmtplib

logger = logging.getLogger("mail")

type MessageRole = Literal["user", "agent"]


class MailThreadContext:
    def __init__(self, *, chat: AgentChatContext, room_address: str):
        self.chat = chat
        self.room_address = room_address


def create_reply_email_message(
    *, message: dict, from_address: str, body: str
) -> EmailMessage:
    subject: str = message.get("subject")

    if not subject.startswith("RE:"):
        subject = "RE: " + subject

    _, addr = email.utils.parseaddr(from_address)
    domain = addr.split("@")[-1].lower()
    id = f"<{uuid.uuid4()}@{domain}>"

    msg = EmailMessage()
    msg["Message-ID"] = id
    msg["Subject"] = subject
    msg["From"] = from_address
    msg["To"] = message.get("reply_to")
    msg["In-Reply-To"] = message.get("id")
    msg.set_content(body)

    return msg


def message_to_json(*, message: EmailMessage, role: MessageRole):
    body_part = message.get_body(
        ("plain", "html")
    )  # returns the “best” part :contentReference[oaicite:0]{index=0}
    if body_part:
        body = body_part.get_content()
    else:  # simple, non-MIME message
        body = message.get_content()

    id = message.get("Message-ID")
    if id is None:
        mfrom = message.get("From")
        _, addr = email.utils.parseaddr(mfrom)
        domain = addr.split("@")[-1].lower()
        id = f"{uuid.uuid4()}@{domain}"

    return {
        "id": id,
        "in_reply_to": message.get("In-Reply-To"),
        "reply_to": message.get("Reply-To", message.get("From")),
        "references": message.get("References"),
        "from": message.get("From"),
        "to": message.get_all("To"),
        "subject": message.get("Subject"),
        "body": body,
        "attachments": [],
        "role": role,
    }


async def load_message(*, room: RoomClient, message_id: str) -> dict | None:
    messages = await room.database.search(table="emails", where={"id": message_id})

    if len(messages) == 0:
        return None

    return json.loads(messages[0]["json"])


async def save_email_message(
    *, room: RoomClient, content: bytes, role: MessageRole
) -> dict:
    message = message_from_bytes(content, policy=default)

    now = datetime.now(timezone.utc)

    folder_path = (
        now.strftime("%Y/%m/%d")
        + "/"
        + now.strftime("%H/%M/%S")
        + "/"
        + secrets.token_hex(3)
    )

    queued_message = message_to_json(message=message, role=role)
    message_id = queued_message["id"]

    queued_message["role"] = role

    queued_message["path"] = f".emails/{message_id}/message.json"

    for part in (
        message.iter_attachments()
    ):  # ↔ only the “real” attachments :contentReference[oaicite:0]{index=0}
        fname = (
            part.get_filename() or "attachment.bin"
        )  # RFC 2183 filename, if any :contentReference[oaicite:1]{index=1}

        # get_content() auto-decodes transfer-encodings; returns
        # *str* for text/*, *bytes* for everything else :contentReference[oaicite:2]{index=2}
        data = part.get_content()

        # make sure we write binary data
        bin_data = (
            data.encode(part.get_content_charset("utf-8"))
            if isinstance(data, str)
            else data
        )

        path = f".emails/{folder_path}/attachments/{fname}"
        handle = await room.storage.open(path=path)
        try:
            logger.info(f"writing content to {path}")
            await room.storage.write(handle=handle, data=bin_data)
        finally:
            await room.storage.close(handle=handle)

        queued_message["attachments"].append(path)

    logger.info(f"received mail, {queued_message}")

    # write email
    path = f".emails/{folder_path}/message.eml"
    handle = await room.storage.open(path=path)
    try:
        logger.info(f"writing source message.eml to {path}")
        await room.storage.write(handle=handle, data=content)
    finally:
        await room.storage.close(handle=handle)

    path = f".emails/{folder_path}/message.json"
    handle = await room.storage.open(path=path)
    try:
        logger.info(f"writing source message.json to {path}")
        await room.storage.write(
            handle=handle, data=json.dumps(queued_message, indent=4).encode("utf-8")
        )
    finally:
        await room.storage.close(handle=handle)

    # create email table if it doesn't exist
    tables = await room.database.list_tables()

    if "emails" not in tables:
        await room.database.create_table_with_schema(
            name="emails",
            schema={"id": TextDataType(), "json": TextDataType()},
            mode="create_if_not_exists",
        )

        await room.database.create_scalar_index(table="emails", column="id")

    await room.database.insert(
        table="emails", records=[{"id": message_id, "json": json.dumps(queued_message)}]
    )

    return queued_message


async def load_thread(*, room: RoomClient, message: dict, thread: list[dict]):
    in_reply_to = message.get("in_reply_to", None)
    if in_reply_to is not None:
        source = await load_message(room=room, message_id=in_reply_to)

        if source is not None:
            thread.insert(0, source)

            await load_thread(room=room, message=source, thread=thread)

        else:
            logger.warning(f"message not found {in_reply_to}")


class SmtpConfiguration:
    def __init__(
        self,
        username: Optional[str] = None,
        password: Optional[str] = None,
        port: Optional[int] = None,
        hostname: Optional[str] = None,
    ):
        if username is None:
            username = os.getenv("SMTP_USERNAME")

        if password is None:
            password = os.getenv("SMTP_PASSWORD")

        if port is None:
            port = int(os.getenv("SMTP_PORT", "587"))

        if hostname is None:
            hostname = os.getenv("SMTP_HOSTNAME")

        self.username = username
        self.password = password
        self.port = port
        self.hostname = hostname


class MailWorker(Worker):
    def __init__(
        self,
        *,
        queue: str = "email",
        name,
        title=None,
        description=None,
        requires=None,
        llm_adapter,
        tool_adapter=None,
        toolkits=None,
        rules=None,
        domain: str = os.getenv("MESHAGENT_MAIL_DOMAIN", "mail.meshagent.com"),
        smtp: Optional[SmtpConfiguration] = None,
    ):
        if smtp is None:
            smtp = SmtpConfiguration()

        self._domain = domain
        self._smtp = smtp
        super().__init__(
            queue=queue,
            name=name,
            title=title,
            description=description,
            requires=requires,
            llm_adapter=llm_adapter,
            tool_adapter=tool_adapter,
            toolkits=toolkits,
            rules=rules,
        )

    async def start(self, *, room):
        await super().start(room=room)

        token = ParticipantToken.from_jwt(room.protocol.token, validate=False)
        self._email_address = room_address(
            project_id=token.project_id, room_name=room.room_name, domain=self._domain
        )

    async def append_message_context(self, *, room, message, chat_context):
        thread = [message]

        await load_thread(room=room, message=message, thread=thread)

        for msg in thread:
            if msg["role"] == "agent":
                chat_context.append_assistant_message(json.dumps(msg))

            else:
                chat_context.append_user_message(json.dumps(msg))

        # TODO: load previous messages
        return await super().append_message_context(
            room=room, message=message, chat_context=chat_context
        )

    async def process_message(self, *, chat_context, room, message, toolkits):
        thread_context = MailThreadContext(chat=chat_context, room_address=message)
        toolkits = [
            *toolkits,
            *await self.get_thread_toolkits(thread_context=thread_context),
        ]

        logger.info(f"processing message {message}")
        reply = await super().process_message(
            chat_context=thread_context.chat,
            room=room,
            message=message,
            toolkits=toolkits,
        )
        return await self.send_reply_message(room=room, message=message, reply=reply)

    async def send_reply_message(self, *, room: RoomClient, message: dict, reply: str):
        msg = create_reply_email_message(
            message=message, from_address=self._email_address, body=reply
        )

        reply_msg_dict = await save_email_message(
            room=room, content=msg.as_bytes(), role="agent"
        )

        logger.info(f"replying with message {reply_msg_dict}")

        username = self._smtp.username
        if username is None:
            username = self.room.local_participant.get_attribute("name")

        password = self._smtp.password
        if password is None:
            password = self.room.protocol.token

        hostname = self._smtp.hostname
        if hostname is None:
            hostname = self._domain

        port = self._smtp.port

        logger.info(f"using smtp {username}@{hostname}:{port}")

        await aiosmtplib.send(
            msg,
            hostname=hostname,
            port=port,
            username=username,
            password=password,
        )

    async def get_thread_toolkits(
        self,
        *,
        thread_context: MailThreadContext,
    ) -> list[Toolkit]:
        toolkits = await self.get_required_toolkits(
            context=ToolContext(
                room=self.room,
                caller=self.room.local_participant,
                caller_context={"chat": thread_context.chat.to_json()},
            )
        )

        return [*self._toolkits, *toolkits]


def base36encode(number: int):
    if not isinstance(number, int):
        raise TypeError("number must be an integer")
    if number < 0:
        raise ValueError("number must be non-negative")

    alphabet = "0123456789abcdefghijklmnopqrstuvwxyz"
    if number == 0:
        return "0"
    base36 = ""
    while number:
        number, i = divmod(number, 36)
        base36 = alphabet[i] + base36
    return base36


def compress_uuid(guid_string: str):
    guid_int = int(guid_string.replace("-", ""), 16)
    return base36encode(guid_int)


def base36decode(number_str: str) -> int:
    """Decode a base36-encoded string into an integer."""
    alphabet = "0123456789abcdefghijklmnopqrstuvwxyz"
    base = 36
    number = 0
    for char in number_str:
        try:
            value = alphabet.index(char)
        except ValueError:
            raise ValueError(f"Invalid character '{char}' for base36 encoding")
        number = number * base + value
    return number


def decompress_uuid(compressed_uuid: str) -> str:
    """
    Reverse the compressed UUID to its standard 36-character UUID format.

    Args:
        compressed_uuid: A base36 string that represents a UUID compressed from its standard form.

    Returns:
        A string in the UUID format (8-4-4-4-12 hexadecimal characters).
    """
    # Decode the base36 string back to the original integer.
    guid_int = base36decode(compressed_uuid)

    # Convert the integer into a 32-digit hexadecimal string with leading zeros.
    hex_str = f"{guid_int:032x}"

    # Reinsert dashes to match the standard UUID format: 8-4-4-4-12.
    standard_uuid = f"{hex_str[0:8]}-{hex_str[8:12]}-{hex_str[12:16]}-{hex_str[16:20]}-{hex_str[20:32]}"
    return standard_uuid


def room_address(
    *,
    project_id: str,
    room_name: str,
    domain: str = os.getenv("MESHAGENT_MAIL_DOMAIN", "mail.meshagent.com"),
):
    return f"{compress_uuid(project_id)}+{room_name}@{domain}"



from .agent import SingleRoomAgent, AgentChatContext
from meshagent.api.chan import Chan
from meshagent.api import (
    RoomMessage,
    RoomClient,
    RemoteParticipant,
    RequiredSchema,
    Requirement,
    Element,
    MeshDocument,
)
from meshagent.tools import Toolkit, ToolContext
from .adapter import LLMAdapter, ToolResponseAdapter
from meshagent.openai.tools.responses_adapter import ImageGenerationTool, LocalShellTool
import asyncio
from typing import Optional
import logging
from meshagent.tools import MultiToolkit
import uuid
import datetime
from typing import Literal
import base64
from openai.types.responses import ResponseStreamEvent
from asyncio import CancelledError

from opentelemetry import trace
import shlex

tracer = trace.get_tracer("meshagent.chatbot")

logger = logging.getLogger("chat")


class ChatBotThreadLocalShellTool(LocalShellTool):
    def __init__(self, *, thread_context: "ChatThreadContext"):
        super().__init__()
        self.thread_context = thread_context

    async def execute_shell_command(
        self,
        context,
        *,
        command,
        env,
        type,
        timeout_ms=None,
        user=None,
        working_directory=None,
    ):
        messages = None

        for prop in self.thread_context.thread.root.get_children():
            if prop.tag_name == "messages":
                messages = prop

        exec_element = messages.append_child(
            tag_name="exec",
            attributes={"command": shlex.join(command), "pwd": working_directory},
        )

        result = await super().execute_shell_command(
            context,
            command=command,
            env=env,
            type=type,
            timeout_ms=timeout_ms,
            user=user,
            working_directory=working_directory,
        )

        exec_element.set_attribute("result", result)

        return result


class ChatBotThreadOpenAIImageGenerationTool(ImageGenerationTool):
    def __init__(
        self,
        *,
        background: Literal["transparent", "opaque", "auto"] = None,
        input_image_mask_url: Optional[str] = None,
        model: Optional[str] = None,
        moderation: Optional[str] = None,
        output_compression: Optional[int] = None,
        output_format: Optional[Literal["png", "webp", "jpeg"]] = None,
        partial_images: Optional[int] = None,
        quality: Optional[Literal["auto", "low", "medium", "high"]] = None,
        size: Optional[Literal["1024x1024", "1024x1536", "1536x1024", "auto"]] = None,
        thread_context: "ChatThreadContext",
    ):
        super().__init__(
            background=background,
            input_image_mask_url=input_image_mask_url,
            model=model,
            moderation=moderation,
            output_compression=output_compression,
            output_format=output_format,
            partial_images=partial_images,
            quality=quality,
            size=size,
        )

        self.thread_context = thread_context

    async def on_image_generation_partial(
        self,
        context,
        *,
        item_id,
        output_index,
        sequence_number,
        type,
        partial_image_b64,
        partial_image_index,
        size,
        quality,
        background,
        output_format,
        **extra,
    ):
        output_format = self.output_format
        if output_format is None:
            output_format = "png"

        image_name = f"{str(uuid.uuid4())}.{output_format}"

        handle = await context.room.storage.open(path=image_name)
        await context.room.storage.write(
            handle=handle, data=base64.b64decode(partial_image_b64)
        )
        await context.room.storage.close(handle=handle)

        messages = None

        for prop in self.thread_context.thread.root.get_children():
            if prop.tag_name == "messages":
                messages = prop

        for child in messages.get_children():
            if child.get_attribute("id") == item_id:
                for file in child.get_children():
                    file.set_attribute("path", image_name)

                return

        message_element = messages.append_child(
            tag_name="message",
            attributes={
                "id": item_id,
                "text": "",
                "created_at": datetime.datetime.now(datetime.timezone.utc)
                .isoformat()
                .replace("+00:00", "Z"),
                "author_name": context.room.local_participant.get_attribute("name"),
            },
        )
        message_element.append_child(tag_name="file", attributes={"path": image_name})

    async def on_image_generated(
        self,
        context: ToolContext,
        *,
        item_id: str,
        data: bytes,
        status: str,
        size: str,
        quality: str,
        background: str,
        output_format: str,
        **extra,
    ):
        output_format = self.output_format
        if output_format is None:
            output_format = "png"

        image_name = f"{str(uuid.uuid4())}.{output_format}"

        handle = await context.room.storage.open(path=image_name)
        await context.room.storage.write(handle=handle, data=data)
        await context.room.storage.close(handle=handle)

        messages = None

        for prop in self.thread_context.thread.root.get_children():
            if prop.tag_name == "messages":
                messages = prop

        for child in messages.get_children():
            if child.get_attribute("id") == item_id:
                for file in child.get_children():
                    file.set_attribute("path", image_name)

                return

        message_element = messages.append_child(
            tag_name="message",
            attributes={
                "id": item_id,
                "text": "",
                "created_at": datetime.datetime.now(datetime.timezone.utc)
                .isoformat()
                .replace("+00:00", "Z"),
                "author_name": context.room.local_participant.get_attribute("name"),
            },
        )
        message_element.append_child(tag_name="file", attributes={"path": image_name})

        self.thread_context.chat.append_assistant_message(
            f"An image was saved at the path {image_name} and displayed to the user"
        )


def get_thread_participants(
    *, room: RoomClient, thread: MeshDocument
) -> list[RemoteParticipant]:
    results = list[RemoteParticipant]()

    for prop in thread.root.get_children():
        if prop.tag_name == "members":
            for member in prop.get_children():
                for online in room.messaging.get_participants():
                    if online.get_attribute("name") == member.get_attribute("name"):
                        results.append(online)

    return results


class ChatThreadContext:
    def __init__(
        self,
        *,
        chat: AgentChatContext,
        thread: MeshDocument,
        path: str,
        participants: Optional[list[RemoteParticipant]] = None,
    ):
        self.thread = thread
        if participants is None:
            participants = []

        self.participants = participants
        self.chat = chat
        self.path = path


# todo: thread should stop when participant stops?
class ChatBot(SingleRoomAgent):
    def __init__(
        self,
        *,
        name,
        title=None,
        description=None,
        requires: Optional[list[Requirement]] = None,
        llm_adapter: LLMAdapter,
        tool_adapter: Optional[ToolResponseAdapter] = None,
        toolkits: Optional[list[Toolkit]] = None,
        rules: Optional[list[str]] = None,
        auto_greet_message: Optional[str] = None,
        empty_state_title: Optional[str] = None,
        labels: Optional[str] = None,
    ):
        super().__init__(
            name=name,
            title=title,
            description=description,
            requires=requires,
            labels=labels,
        )

        if toolkits is None:
            toolkits = []

        self._llm_adapter = llm_adapter
        self._tool_adapter = tool_adapter

        self._message_channels = dict[str, Chan[RoomMessage]]()

        self._room: RoomClient | None = None
        self._toolkits = toolkits

        if rules is None:
            rules = []

        self._rules = rules
        self._is_typing = dict[str, asyncio.Task]()
        self._auto_greet_message = auto_greet_message

        if empty_state_title is None:
            empty_state_title = "How can I help you?"
        self._empty_state_title = empty_state_title

        self._thread_tasks = dict[str, asyncio.Task]()

    def init_requirements(self, requires: list[Requirement]):
        if requires is None:
            requires = [RequiredSchema(name="thread")]

        else:
            thread_schema = list(
                n
                for n in requires
                if (isinstance(n, RequiredSchema) and n.name == "thread")
            )
            if len(thread_schema) == 0:
                requires.append(RequiredSchema(name="thread"))

    async def _send_and_save_chat(
        self,
        messages: Element,
        path: str,
        to: RemoteParticipant,
        id: str,
        text: str,
        thread_attributes: dict,
    ):
        with tracer.start_as_current_span("chatbot.thread.message") as span:
            span.set_attributes(thread_attributes)
            span.set_attribute("role", "assistant")
            span.set_attribute(
                "from_participant_name",
                self.room.local_participant.get_attribute("name"),
            )
            span.set_attributes({"id": id, "text": text})

            await self.room.messaging.send_message(
                to=to, type="chat", message={"path": path, "text": text}
            )

            messages.append_child(
                tag_name="message",
                attributes={
                    "id": id,
                    "text": text,
                    "created_at": datetime.datetime.now(datetime.timezone.utc)
                    .isoformat()
                    .replace("+00:00", "Z"),
                    "author_name": self.room.local_participant.get_attribute("name"),
                },
            )

    async def greet(
        self,
        *,
        messages: Element,
        path: str,
        chat_context: AgentChatContext,
        participant: RemoteParticipant,
        thread_attributes: dict,
    ):
        if self._auto_greet_message is not None:
            chat_context.append_user_message(self._auto_greet_message)
            await self._send_and_save_chat(
                id=str(uuid.uuid4()),
                to=RemoteParticipant(id=participant.id),
                messages=messages,
                path=path,
                text=self._auto_greet_message,
                thread_attributes=thread_attributes,
            )

    async def get_thread_participants(self, *, thread: MeshDocument):
        return get_thread_participants(room=self._room, thread=thread)

    async def get_thread_toolkits(
        self, *, thread_context: ChatThreadContext, participant: RemoteParticipant
    ) -> list[Toolkit]:
        toolkits = await self.get_required_toolkits(
            context=ToolContext(
                room=self.room,
                caller=participant,
                caller_context={"chat": thread_context.chat.to_json()},
            )
        )
        toaster = None

        for toolkit in toolkits:
            if toolkit.name == "ui":
                for tool in toolkit.tools:
                    if tool.name == "show_toast":
                        toaster = tool

        if toaster is not None:

            def multi_tool(toolkit: Toolkit):
                if toaster in toolkit.tools:
                    return toolkit

                return MultiToolkit(required=[toaster], base_toolkit=toolkit)

            toolkits = list(map(multi_tool, toolkits))

        return [*self._toolkits, *toolkits]

    async def init_chat_context(self) -> AgentChatContext:
        context = self._llm_adapter.create_chat_context()
        context.append_rules(self._rules)
        return context

    async def open_thread(self, *, path: str):
        return await self.room.sync.open(path=path)

    async def close_thread(self, *, path: str):
        return await self.room.sync.close(path=path)

    async def _spawn_thread(self, path: str, messages: Chan[RoomMessage]):
        self.room.developer.log_nowait(
            type="chatbot.thread.started", data={"path": path}
        )
        chat_context = await self.init_chat_context()
        opened = False

        doc_messages = None
        current_file = None
        llm_messages = Chan[ResponseStreamEvent]()
        thread_context = None

        thread_attributes = None

        def done_processing_llm_events(task: asyncio.Task):
            try:
                task.result()
            except CancelledError:
                pass
            except Exception as e:
                logger.error("error sending delta", exc_info=e)

        async def process_llm_events():
            partial = ""
            content_element = None
            context_message = None

            async for evt in llm_messages:
                for participant in self._room.messaging.get_participants():
                    logger.debug(
                        f"sending event {evt.type} to {participant.get_attribute('name')}"
                    )

                    # self.room.messaging.send_message_nowait(to=participant, type="llm.event", message=json.loads(evt.to_json()))

                if evt.type == "response.content_part.added":
                    partial = ""
                    content_element = doc_messages.append_child(
                        tag_name="message",
                        attributes={
                            "text": "",
                            "created_at": datetime.datetime.now(datetime.timezone.utc)
                            .isoformat()
                            .replace("+00:00", "Z"),
                            "author_name": self.room.local_participant.get_attribute(
                                "name"
                            ),
                        },
                    )

                    context_message = {"role": "assistant", "content": ""}
                    chat_context.messages.append(context_message)

                elif evt.type == "response.output_text.delta":
                    partial += evt.delta
                    content_element["text"] = partial
                    context_message["content"] = partial

                elif evt.type == "response.output_text.done":
                    content_element = None

                    with tracer.start_as_current_span("chatbot.thread.message") as span:
                        span.set_attribute(
                            "from_participant_name",
                            self.room.local_participant.get_attribute("name"),
                        )
                        span.set_attribute("role", "assistant")
                        span.set_attributes(thread_attributes)
                        span.set_attributes({"text": evt.text})

        llm_task = asyncio.create_task(process_llm_events())
        llm_task.add_done_callback(done_processing_llm_events)

        thread = None

        try:
            received = None

            while True:
                while True:
                    logger.info(f"waiting for message on thread {path}")
                    received = await messages.recv()
                    logger.info(f"received message on thread {path}: {received.type}")

                    chat_with_participant = None
                    for participant in self._room.messaging.get_participants():
                        if participant.id == received.from_participant_id:
                            chat_with_participant = participant
                            break

                    if chat_with_participant is None:
                        logger.warning(
                            "participant does not have messaging enabled, skipping message"
                        )
                        continue

                    thread_attributes = {
                        "agent_name": self.name,
                        "agent_participant_id": self.room.local_participant.id,
                        "agent_participant_name": self.room.local_participant.get_attribute(
                            "name"
                        ),
                        "remote_participant_id": chat_with_participant.id,
                        "remote_participant_name": chat_with_participant.get_attribute(
                            "name"
                        ),
                        "path": path,
                    }

                    if current_file != chat_with_participant.get_attribute(
                        "current_file"
                    ):
                        logger.info(
                            f"participant is now looking at {chat_with_participant.get_attribute('current_file')}"
                        )
                        current_file = chat_with_participant.get_attribute(
                            "current_file"
                        )

                    if current_file is not None:
                        chat_context.append_assistant_message(
                            message=f"the user is currently viewing the file at the path: {current_file}"
                        )

                    elif current_file is not None:
                        chat_context.append_assistant_message(
                            message="the user is not current viewing any files"
                        )

                    if thread is None:
                        with tracer.start_as_current_span(
                            "chatbot.thread.open"
                        ) as span:
                            span.set_attributes(thread_attributes)

                            thread = await self.open_thread(path=path)

                            for prop in thread.root.get_children():
                                if prop.tag_name == "messages":
                                    doc_messages = prop

                                    for element in doc_messages.get_children():
                                        if (
                                            isinstance(element, Element)
                                            and element.tag_name == "message"
                                        ):
                                            msg = element["text"]
                                            if (
                                                element["author_name"]
                                                == self.room.local_participant.get_attribute(
                                                    "name"
                                                )
                                            ):
                                                chat_context.append_assistant_message(
                                                    msg
                                                )
                                            else:
                                                chat_context.append_user_message(msg)

                                            for child in element.get_children():
                                                if child.tag_name == "file":
                                                    chat_context.append_assistant_message(
                                                        f"the user attached a file with the path '{child.get_attribute('path')}'"
                                                    )

                            if doc_messages is None:
                                raise Exception("thread was not properly initialized")

                    if received.type == "opened":
                        if not opened:
                            opened = True

                            await self.greet(
                                path=path,
                                chat_context=chat_context,
                                participant=chat_with_participant,
                                messages=doc_messages,
                                thread_attributes=thread_attributes,
                            )

                    if received.type == "chat":
                        if thread is None:
                            self.room.developer.log_nowait(
                                type="thread is not open", data={}
                            )
                            break

                        for participant in get_thread_participants(
                            room=self._room, thread=thread
                        ):
                            # TODO: async gather
                            self._room.messaging.send_message_nowait(
                                to=participant,
                                type="thinking",
                                message={"thinking": True, "path": path},
                            )

                        if chat_with_participant.id == received.from_participant_id:
                            self.room.developer.log_nowait(
                                type="llm.message",
                                data={
                                    "context": chat_context.id,
                                    "participant_id": self.room.local_participant.id,
                                    "participant_name": self.room.local_participant.get_attribute(
                                        "name"
                                    ),
                                    "message": {
                                        "content": {
                                            "role": "user",
                                            "text": received.message["text"],
                                        }
                                    },
                                },
                            )

                            attachments = received.message.get("attachments", [])
                            text = received.message["text"]

                            for attachment in attachments:
                                chat_context.append_assistant_message(
                                    message=f"the user attached a file at the path '{attachment['path']}'"
                                )

                            chat_context.append_user_message(message=text)

                        # if user is typing, wait for typing to stop
                        while True:
                            if chat_with_participant.id not in self._is_typing:
                                break

                            await asyncio.sleep(0.5)

                        if messages.empty():
                            break

                if received is not None:
                    with tracer.start_as_current_span("chatbot.thread.message") as span:
                        span.set_attributes(thread_attributes)
                        span.set_attribute("role", "user")
                        span.set_attribute(
                            "from_participant_name",
                            chat_with_participant.get_attribute("name"),
                        )

                        attachments = received.message.get("attachments", [])
                        span.set_attribute("attachments", attachments)

                        text = received.message["text"]
                        span.set_attributes({"text": text})

                        try:
                            if thread_context is None:
                                thread_context = ChatThreadContext(
                                    path=path,
                                    chat=chat_context,
                                    thread=thread,
                                    participants=get_thread_participants(
                                        room=self.room, thread=thread
                                    ),
                                )
                            else:
                                thread_context.participants = get_thread_participants(
                                    room=self.room, thread=thread
                                )

                            def handle_event(evt):
                                llm_messages.send_nowait(evt)

                            with tracer.start_as_current_span("chatbot.llm") as span:
                                try:
                                    with tracer.start_as_current_span(
                                        "get_thread_toolkits"
                                    ) as span:
                                        thread_toolkits = (
                                            await self.get_thread_toolkits(
                                                thread_context=thread_context,
                                                participant=participant,
                                            )
                                        )

                                    await self._llm_adapter.next(
                                        context=chat_context,
                                        room=self._room,
                                        toolkits=thread_toolkits,
                                        tool_adapter=self._tool_adapter,
                                        event_handler=handle_event,
                                    )
                                except Exception as e:
                                    logger.error("An error was encountered", exc_info=e)
                                    await self._send_and_save_chat(
                                        messages=doc_messages,
                                        to=chat_with_participant,
                                        path=path,
                                        id=str(uuid.uuid4()),
                                        text="There was an error while communicating with the LLM. Please try again later.",
                                        thread_attributes=thread_attributes,
                                    )

                        finally:
                            for participant in get_thread_participants(
                                room=self._room, thread=thread
                            ):
                                # TODO: async gather
                                self._room.messaging.send_message_nowait(
                                    to=participant,
                                    type="thinking",
                                    message={"thinking": False, "path": path},
                                )

        finally:
            llm_messages.close()

            if self.room is not None:
                logger.info("thread was ended {path}")
                self.room.developer.log_nowait(
                    type="chatbot.thread.ended", data={"path": path}
                )

                if thread is not None:
                    await self.close_thread(path=path)

    def _get_message_channel(self, key: str) -> Chan[RoomMessage]:
        if key not in self._message_channels:
            chan = Chan[RoomMessage]()
            self._message_channels[key] = chan

        chan = self._message_channels[key]

        return chan

    async def stop(self):
        await super().stop()

        for thread in self._thread_tasks.values():
            thread.cancel()

        self._thread_tasks.clear()

    async def start(self, *, room):
        await super().start(room=room)

        logger.debug("Starting chatbot")

        await self.room.local_participant.set_attribute(
            "empty_state_title", self._empty_state_title
        )

        def on_message(message: RoomMessage):
            if message.type == "chat" or message.type == "opened":
                path = message.message["path"]

                messages = self._get_message_channel(path)

                logger.info(
                    f"queued incoming message for thread {path}: {message.type}"
                )

                messages.send_nowait(message)

                if path not in self._thread_tasks or self._thread_tasks[path].done():

                    def thread_done(task: asyncio.Task):
                        self._message_channels.pop(path)
                        try:
                            task.result()
                        except CancelledError:
                            pass
                        except Exception as e:
                            logger.error(
                                f"The chat thread ended with an error {e}", exc_info=e
                            )

                    logger.info(f"spawning chat thread for {path}")
                    task = asyncio.create_task(
                        self._spawn_thread(messages=messages, path=path)
                    )
                    task.add_done_callback(thread_done)

                    self._thread_tasks[path] = task

            elif message.type == "typing":

                def callback(task: asyncio.Task):
                    try:
                        task.result()
                    except CancelledError:
                        pass
                    except Exception:
                        pass

                async def remove_timeout(id: str):
                    await asyncio.sleep(1)
                    self._is_typing.pop(id)

                if message.from_participant_id in self._is_typing:
                    self._is_typing[message.from_participant_id].cancel()

                timeout = asyncio.create_task(
                    remove_timeout(id=message.from_participant_id)
                )
                timeout.add_done_callback(callback)

                self._is_typing[message.from_participant_id] = timeout

        room.messaging.on("message", on_message)

        if self._auto_greet_message is not None:

            def on_participant_added(participant: RemoteParticipant):
                # will spawn the initial thread
                self._get_message_channel(participant.id)

            room.messaging.on("participant_added", on_participant_added)

        logger.debug("Enabling chatbot messaging")
        await room.messaging.enable()



from meshagent.api.schema import MeshSchema, ElementType, ChildProperty, ValueProperty

thread_schema = MeshSchema(
    root_tag_name="thread",
    elements=[
        ElementType(
            tag_name="thread",
            description="a thread of messages",
            properties=[
                ValueProperty(
                    name="name", description="name of the tread", type="string"
                ),
                ChildProperty(
                    name="properties",
                    description="the messages in the thread",
                    ordered=True,
                    child_tag_names=["members", "messages"],
                ),
            ],
        ),
        ElementType(
            tag_name="members",
            description="the members of this thread",
            properties=[
                ChildProperty(
                    name="items",
                    child_tag_names=["member"],
                    description="the messages in this thread",
                )
            ],
        ),
        ElementType(
            tag_name="messages",
            description="the messages of this thread",
            properties=[
                ChildProperty(
                    name="items",
                    child_tag_names=["message", "exec"],
                    description="the messages in this thread",
                )
            ],
        ),
        ElementType(
            tag_name="member",
            description="a member of this thread",
            properties=[
                ValueProperty(
                    name="name", description="the name of the member", type="string"
                ),
                ValueProperty(
                    name="type",
                    description="the type of member",
                    type="string",
                    enum=["user", "agent"],
                ),
            ],
        ),
        ElementType(
            tag_name="file",
            description="a file attachment",
            properties=[
                ValueProperty(
                    name="path",
                    description="the path of the file in the room",
                    type="string",
                ),
            ],
        ),
        ElementType(
            tag_name="exec",
            description="a command execution",
            properties=[
                ValueProperty(
                    name="command",
                    description="a command that was executed",
                    type="string",
                ),
                ValueProperty(
                    name="result",
                    description="the result of the command",
                    type="string",
                ),
                ValueProperty(
                    name="pwd",
                    description="the working directory the command was executed in",
                    type="string",
                ),
            ],
        ),
        ElementType(
            tag_name="message",
            description="a message sent in the conversation",
            properties=[
                ValueProperty(
                    name="id", description="the id of the message", type="string"
                ),
                ValueProperty(
                    name="text", description="the text of the message", type="string"
                ),
                ValueProperty(
                    name="created_at",
                    description="the date that the message was sent in ISO format",
                    type="string",
                ),
                ValueProperty(
                    name="author_name",
                    description="the name of the author of the post",
                    type="string",
                ),
                ValueProperty(
                    name="author_ref",
                    description="a reference to author identity in another system",
                    type="string",
                ),
                ChildProperty(
                    name="attachments",
                    child_tag_names=["file"],
                    description="a list of message attachments",
                ),
            ],
        ),
    ],
)



import logging

from meshagent.api import RoomMessage
from meshagent.api.webhooks import WebhookServer, CallEvent
from meshagent.api import WebSocketClientProtocol
from meshagent.api.room_server_client import RoomClient
from meshagent.agents import SingleRoomAgent
from aiohttp import web
import asyncio

from typing import Callable, Optional

from .agent import TaskRunner

logger = logging.getLogger("hosting")


class RemoteTaskRunnerServer[T: TaskRunner](WebhookServer):
    def __init__(
        self,
        *,
        cls: Optional[T] = None,
        path: Optional[str] = None,
        app: Optional[web.Application] = None,
        host=None,
        port=None,
        webhook_secret=None,
        create_agent: Optional[Callable[[dict], TaskRunner]] = None,
        validate_webhook_secret: Optional[bool] = None,
    ):
        super().__init__(
            path=path,
            app=app,
            host=host,
            port=port,
            webhook_secret=webhook_secret,
            validate_webhook_secret=validate_webhook_secret,
        )

        if create_agent is None:

            def default_create_agent(arguments: dict) -> TaskRunner:
                return cls(**arguments)

            create_agent = default_create_agent

        self._create_agent = create_agent

    async def _spawn(
        self,
        *,
        room_name: str,
        room_url: str,
        token: str,
        arguments: Optional[dict] = None,
    ):
        agent = self._create_agent(arguments=arguments)

        async def run():
            async with RoomClient(
                protocol=WebSocketClientProtocol(url=room_url, token=token)
            ) as room:
                dismissed = asyncio.Future()

                def on_message(message: RoomMessage):
                    if message.type == "dismiss":
                        logger.info(
                            f"dismissed task runner by {message.from_participant_id}"
                        )
                        dismissed.set_result(True)

                room.messaging.on("message", on_message)

                await agent.start(room=room)

                done, pending = await asyncio.wait(
                    [dismissed, asyncio.ensure_future(room.protocol.wait_for_close())],
                    return_when=asyncio.FIRST_COMPLETED,
                )

                await agent.stop()

        def on_done(task: asyncio.Task):
            try:
                task.result()
            except Exception as e:
                logger.error("agent encountered an error", exc_info=e)

        task = asyncio.create_task(run())
        task.add_done_callback(on_done)

    async def on_call(self, event: CallEvent):
        await self._spawn(
            room_name=event.room_name,
            room_url=event.room_url,
            token=event.token,
            arguments=event.arguments,
        )


class RemoteAgentServer[T: SingleRoomAgent](WebhookServer):
    def __init__(
        self,
        *,
        cls: Optional[T] = None,
        path: Optional[str] = None,
        app: Optional[web.Application] = None,
        host=None,
        port=None,
        webhook_secret=None,
        create_agent: Optional[Callable[[dict], SingleRoomAgent]] = None,
        validate_webhook_secret: Optional[bool] = None,
    ):
        super().__init__(
            path=path,
            app=app,
            host=host,
            port=port,
            webhook_secret=webhook_secret,
            validate_webhook_secret=validate_webhook_secret,
        )

        if create_agent is None:

            def default_create_agent(arguments: dict) -> SingleRoomAgent:
                return cls(**arguments)

            create_agent = default_create_agent

        self._create_agent = create_agent

    async def _spawn(
        self,
        *,
        room_name: str,
        room_url: str,
        token: str,
        arguments: Optional[dict] = None,
    ):
        logger.info(
            f"spawning agent on room: {room_name} url: {room_url} arguments: {arguments}"
        )
        agent = self._create_agent(arguments=arguments)

        async def run():
            async with RoomClient(
                protocol=WebSocketClientProtocol(url=room_url, token=token)
            ) as room:
                dismissed = asyncio.Future()

                def on_message(message: RoomMessage):
                    if message.type == "dismiss":
                        logger.info(f"dismissed agent by {message.from_participant_id}")
                        dismissed.set_result(True)

                room.messaging.on("message", on_message)

                await agent.start(room=room)

                done, pending = await asyncio.wait(
                    [dismissed, asyncio.ensure_future(room.protocol.wait_for_close())],
                    return_when=asyncio.FIRST_COMPLETED,
                )

                await agent.stop()

        def on_done(task: asyncio.Task):
            try:
                task.result()
            except Exception as e:
                logger.error("agent encountered an error", exc_info=e)

        task = asyncio.create_task(run())
        task.add_done_callback(on_done)

    async def on_call(self, event: CallEvent):
        await self._spawn(
            room_name=event.room_name,
            room_url=event.room_url,
            token=event.token,
            arguments=event.arguments,
        )
